@@ -10,9 +10,10 @@ from enum import Enum
 from enum import StrEnum
 from types import UnionType
 from typing import ClassVar
-from typing import Optional, Type, Any
+from typing import Optional, Type, Any, Iterable, Callable, Sized
 
 from apispec import APISpec
+from dict_caster.extras import first
 from http_tools import Answer
 from init_helpers.dict_to_dataclass import NoValue
 
@@ -142,6 +143,22 @@ class OpenApiWrapper:
         self.context = context
         self.endpoints = []
         self.dataclass_type_to_name: dict[type[DataclassProtocol] | UnionType, str] = {}
+        self.origin_type_to_factory: dict[type | UnionType, Callable[[APISpec, Iterable[type | UnionType]], dict]] = {
+            typing.Union: self._get_union_description,
+            types.UnionType: self._get_union_description,
+            list: self._get_list_description,
+            dict: self._get_dict_description,
+        }
+        self.primitive_type_to_description: dict[type, dict] = {
+            bool: {"type": "boolean"},  # MUST be before int
+            int: {"type": "integer"},
+            float: {"type": "number"},
+            bytes: {"type": "string", "format": "binary"},
+            str: {"type": "string"},
+            Decimal: {"type": "string", 'pattern': r'^\d*\.?\d*$'},
+            type(None): {"type": "null"},
+            None: {"type": "null"},
+        }
 
     def register_endpoint(self, endpoint: Endpoint) -> None:
         self.endpoints.append(endpoint)
@@ -204,69 +221,70 @@ class OpenApiWrapper:
         return result
 
     @functools.cache
-    def _add_schema_to_spec(self, spec: APISpec, schema: type | UnionType, default: Any = NoValue) -> dict:
+    def _get_union_description(self, spec: APISpec, type_args: tuple[type | UnionType]) -> dict:
+        assert type_args and len(type_args) >= 1, f'bad union element type: {type_args}'
+        return {"anyOf": [self._add_schema_to_spec(spec, arg) for arg in type_args]}
+
+    @functools.cache
+    def _get_list_description(self, spec: APISpec, type_args: tuple[type | UnionType] | None) -> dict:
+        if type_args:
+            assert type_args and len(type_args) == 1, f'bad list element type: {type_args}'
+            return {"type": "array", "items": self._add_schema_to_spec(spec, type_args[0])}
+        else:
+            return {"type": "array"}
+
+    def _get_dict_description(self, spec: APISpec, type_args: tuple[type | UnionType] | None) -> dict:
+        if type_args:
+            assert len(type_args) == 2, f'bad dict item type: {type_args}'
+            return {"type": "object", "additionalProperties": self._add_schema_to_spec(spec, type_args[1])}
+        else:
+            return {"type": "object"}
+
+    def _get_dataclass_description(self, spec: APISpec, schema: DataclassProtocol) -> dict:
+        return {"type": "object", "properties": {
+            field_.name: self._add_schema_to_spec(spec, field_.type)
+            for field_ in dataclasses.fields(schema) if field_.repr
+        }}
+
+    @functools.cache
+    def _get_generic_description(
+            self, spec: APISpec, origin_type: type | UnionType, type_args: tuple[type | UnionType] | None) -> dict:
+        if factory := self.origin_type_to_factory.get(origin_type):
+            return factory(spec, type_args)
+        else:
+            raise TypeError(f"Unknown origin_type: {origin_type}")
+
+    @classmethod
+    def _save_schema_to_spec(cls, spec: APISpec, description: dict, name: str) -> dict:
+        logger.debug(f'save schema {name!r}: {description}')
+        spec.components.schema(component_id=name, component=description)
+        return spec.components.get_ref("schema", name)
+
+    @functools.cache
+    def _get_primitive_description(self, type_: type) -> dict:
+        for primitive_type, description in self.primitive_type_to_description.items():
+            if issubclass(type_, primitive_type):
+                return description
+        raise TypeError(f"Unknown type: {type_}")
+
+    @functools.cache
+    def _add_schema_to_spec(
+            self, spec: APISpec, schema: type | UnionType, default: Any = NoValue, base: dict | None = None
+    ) -> dict:
+        base = base or {}
         logger.debug('_add_schema_to_spec: %s', schema)
 
         origin_type = typing.get_origin(schema)
-        type_args = typing.get_args(schema)
-        if origin_type in (typing.Union, types.UnionType):
-            description = {
-                "anyOf": [self._add_schema_to_spec(spec, arg) for arg in type_args],
-            }
-        elif origin_type is list:
-            assert type_args and len(type_args) == 1, f'bad list element type: {type_args}'
-            description = {
-                "type": "array",
-                "items": self._add_schema_to_spec(spec, type_args[0])
-            }
-        elif origin_type is dict:
-            assert type_args and len(type_args) == 2, f'bad dict item types: {type_args}'
-            description = {
-                "type": "object",
-                "additionalProperties": self._add_schema_to_spec(spec, type_args[1])
-            }
-        elif origin_type is not None:
-            raise TypeError(f"Unknown origin_type: {schema}")
-
+        if origin_type is not None:
+            description = self._get_generic_description(spec, origin_type, typing.get_args(schema))
+        elif schema in self.origin_type_to_factory:
+            description = self._get_generic_description(spec, schema, None)
         elif dataclasses.is_dataclass(schema):
-            description = {
-                "type": "object",
-                "properties": {}
-            }
-            # noinspection PyDataclass
-            for field_ in dataclasses.fields(schema):
-                if field_.repr:
-                    description["properties"][field_.name] = self._add_schema_to_spec(spec, field_.type)
-
+            description = self._get_dataclass_description(spec, schema)
             name = self.dataclass_type_to_name.get(schema, f"schema_{schema.__name__}")
-            logger.debug('schema: %s, %s, %s', name, schema, list(description["properties"]))
-            spec.components.schema(component_id=name, component=description)
-            return spec.components.get_ref("schema", name)
-        elif schema is bool:  # MUST be before int
-            description = {"type": "boolean"}
-        elif issubclass(schema, int):
-            description = {
-                "type": "integer",
-                # "format": "int32"
-            }
-        elif issubclass(schema, float):
-            description = {
-                "type": "number",
-                # "format": "int32"
-            }
-        elif issubclass(schema, bytes):
-            description = {
-                "type": "string",
-                "format": "binary"
-            }
-        elif issubclass(schema, str):
-            description = {"type": "string"}
-        elif issubclass(schema, Decimal):
-            description = {"type": "string", 'pattern': r'^\d*\.?\d*$'}
-        elif schema in (None, type(None)):
-            return {"type": "null"}
+            return self._save_schema_to_spec(spec, description, name)
         else:
-            raise TypeError(f"Unknown type: {schema}")
+            description = self._get_primitive_description(schema)
 
         if isinstance(schema, type) and issubclass(schema, Enum):
             description['enum'] = [e.value for e in schema]
@@ -276,7 +294,7 @@ class OpenApiWrapper:
         # if default == NoValue:  # TODO: think about it, should we place "required" inside schema properties or not?
         #     description['required'] = True
 
-        return description
+        return base | description
 
     @functools.cache
     def _add_answer_to_spec(self, spec: APISpec, answer: Type[Answer]) -> str:
